# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/utils/01_utils.ipynb.

# %% auto 0
__all__ = ['combine_dict', 'convert_str_none_to_none', 'load_agent_function', 'merge_with_namespace',
           'delete_lag_window_from_config', 'get_git_revision_hash', 'get_version', 'init_pipeline',
           'prepare_hyperparameter_opt', 'prepare_lag_window', 'prepare_calculation_function',
           'get_actions_per_product', 'track_normalization', 'create_env_agent']

# %% ../nbs/utils/01_utils.ipynb 4
# General libraries:
# import numpy as np
# import pickle
import yaml
# import os
# from tqdm import tqdm, trange
import wandb
import datetime
import torch
import pkg_resources
# import sys,traceback

# %% ../nbs/utils/01_utils.ipynb 6
def combine_dict(dict1, dict2):

    """
    Combines two dictionaries. Raises error if there are overlapping keys.

    Args:
        dict1 (dict): First dictionary
        dict2 (dict): Second dictionary

    Returns:
        dict: Combined dictionary
    """

    overlapping_keys = set(dict1.keys()).intersection(set(dict2.keys()))
    if overlapping_keys:
        raise ValueError(f"Overlapping keys detected: {overlapping_keys}")
    return {**dict1, **dict2}

def convert_str_none_to_none(d):
    """Recursively convert all string "None" in a dictionary to Python's None."""
    for key, value in d.items():
        if value == "None":
            d[key] = None
        elif isinstance(value, dict):
            convert_str_none_to_none(value)
    return d

def load_agent_function(agent_name, agent_mapping):

    """
    Load an agent function based on agent_name.
    The function must be in the same folder as the main file that calls this function.

    Args:
        agent_name (str): Name of the agent
        agent_mapping (dict): Dictionary mapping agent names to agent functions

    Returns:    
        function: Agent function

    """

    if agent_name not in agent_mapping:
        raise ValueError(f"Unknown agent name: {agent_name}")

    module_name, function_name = agent_mapping[agent_name].rsplit('.', 1)
    module = __import__(module_name, fromlist=[function_name])
    create_agent = getattr(module, function_name)

    return create_agent.create_agent

def merge_with_namespace(target_dict, source_dict, target_dict_name):
    
    """
    Merge source_dict into target_dict, using the keys as namespaces.
    For example, if target_dict_name is "agent", the key "agent-epsilon" in source_dict will be merged into target_dict["epsilon"].
    The function is to merge hyperparameters from a config file with the default hyperparameters from the yaml files

    Args:
        target_dict (dict): Target dictionary
        source_dict (dict): Source dictionary
        target_dict_name (str): Name of the target dictionary

    Returns:
        dict: Merged dictionary


    """
    for namespaced_key, value in source_dict.items():
        keys = namespaced_key.split('-')

        if keys[0] != target_dict_name:
            continue

        keys = keys[1:]

        d = target_dict
    
        # Check if keys already exist in target_dict
        exists = True
        for key in keys:
            if key not in d:
                exists = False
                break
            if isinstance(d[key], dict):
                d = d[key]
                continue
            else:
                break

        # If all keys are present, overwrite the value
        if exists:
            prev = d[key]
            d[key] = value
            print(f"Overwriting in key {namespaced_key} value {prev} with value {value}")
        else:
            # exception if key is not present in target_dict
            print(f"Key {namespaced_key} not found in {target_dict_name}.")
            raise ValueError(f"Key {namespaced_key} not found in {target_dict_name}.")

    return target_dict

def delete_lag_window_from_config(config_agent):

    """
    Delete lag_window from config_agent.
    This is necessary to define lag_window per agent in yaml file,
    but the information is used by the environment, not the agent.
    Befor applying the agent it has to be removed.

    """

    for agent_param_name in config_agent.keys():
        if "lag_window" in config_agent[agent_param_name]:
            del config_agent[agent_param_name]["lag_window"]
        
def get_git_revision_hash(directory):
    """
    Get curret git hash of the repository (e.g., to be logged in wandb).
    """
    return subprocess.check_output(['git', 'rev-parse', 'HEAD'], cwd=directory).decode('ascii').strip()

def get_version(library_name: str) -> str:
    """
    Get the version of a specified library.

    :param library_name: Name of the library
    :return: Version of the library
    """
    
    try:
        version = pkg_resources.get_distribution(library_name).version
        return version
    except pkg_resources.DistributionNotFound:
        print(f"Library '{library_name}' not found.")
        return "Not Installed"
    

# %% ../nbs/utils/01_utils.ipynb 8
def init_pipeline(config_env, config_train, config_agent, project_name):

    torch.set_num_threads(1)

    try:
        sweep_config = wandb.config
    except:
        sweep_config = None

    # Get parames for env and training
    with open(config_env, "r") as f:
        config_env = yaml.safe_load(f)

    with open(config_train, "r") as f:
        config_train = yaml.safe_load(f)
    
    with open(config_agent, "r") as f:
        config_agent = yaml.safe_load(f)

    wandb.init(
        project=project_name,
        name = f"{project_name}_{datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}"
    )

    try:
        rl_om_directory = config_train["rl_om_directory"]
        rl_om_hash = get_git_revision_hash(rl_om_directory)
    except:
        rl_om_hash = get_version("RL_OM")
        
    try:
        mushroom_fork_directory = config_train["mushroom_fork_directory"]
        mushroom_fork_hash = get_git_revision_hash(mushroom_fork_directory)
    except:
        mushroom_fork_hash = get_version("mushroom_rl")

    success = False 
    for experiment_directory in config_train["experiment_directory"]:
        try:
            experiment_hash = get_git_revision_hash(experiment_directory)
            success = True
            break
        except:
            pass
    
    if not success:
        experiment_hash = None
    
    return config_env, config_train, config_agent, sweep_config, rl_om_hash, mushroom_fork_hash, experiment_hash


def prepare_hyperparameter_opt():
        
    if 'hyperparameter_opt' in wandb.config._items:
        hyperparameter_opt = wandb.config._items['hyperparameter_opt']
    else:
        hyperparameter_opt = None
    
    if hyperparameter_opt:
        num_datasets = wandb.config._items["num_datasets"]
        weight_final = wandb.config._items["weight_final"] 
        weight_best = wandb.config._items["weight_best"]
        eval_period = wandb.config._items["eval_period"]
        if "dataset_start" in wandb.config._items:
            dataset_start = wandb.config._items["dataset_start"]
        else:
            dataset_start=1
    else:
        num_datasets = None 
        weight_final = None 
        weight_best = None 
        eval_period = None
        dataset_start = None

    return hyperparameter_opt, num_datasets, weight_final, weight_best, eval_period, dataset_start
    
def prepare_lag_window(config_env, config_agent, agent_param_name):
    if "lag_window" in config_agent[agent_param_name]:
        lag_window = config_agent[agent_param_name]["lag_window"]
        config_env["lag_window"] = lag_window
        delete_lag_window_from_config(config_agent) 
    
    return config_env, config_agent

def prepare_calculation_function(config_env, config_agent, agent_param_name):
    # TODO: Check this functionn works properly
    if "calculation_function" in config_agent[agent_param_name]:
        print("overwriting calculation function")
        config_env["calculations"] = config_agent[agent_param_name]["calculation_function"]
        del config_agent[agent_param_name]["calculation_function"]

    return config_env, config_agent

def get_actions_per_product(config_agent, agent_param_name):
    if "actions_per_product" in config_agent[agent_param_name]:
        actions_per_product = config_agent[agent_param_name]["actions_per_product"]
        del config_agent[agent_param_name]["actions_per_product"]
    else:
        actions_per_product = 1
    
    return actions_per_product

def track_normalization(config_env):
    norm_features = config_env.get("normalize_features", False)
    norm_reward = config_env.get("env_kwargs", {}).get("normalize_reward", False)
    norm_limit_low = config_env["env_kwargs"]["order_limit_low"]
    norm_limit_high = config_env["env_kwargs"]["order_limit_high"]
    
    norm_features = 1 if norm_features else 0
    norm_reward  = 1 if norm_reward else 0

    wandb.config.update({"norm_features": norm_features, "norm_reward": norm_reward, "norm_limit_low": norm_limit_low, "norm_limit_high": norm_limit_high})

def create_env_agent(create_env, config_env, config_train, config_agent, hyperparameter_opt, num_datasets, dataset_start, actions_per_product):
    
    if hyperparameter_opt:

        mdp_train_l = list()
        mdp_val_l = list()
        agent_l= list()
        for i in range(dataset_start-1, dataset_start-1+num_datasets):
            print("creating list of validation sets, not test sets")
            mdp_train, mdp_val, _ = create_env(config_env, create_val=True, use_dataset = i+1, actions_per_product = actions_per_product)

            mdp_train_l.append(mdp_train)
            mdp_val_l.append(mdp_val)

            create_agent = load_agent_function(config_train["agent"], config_agent["AGENTS"])

            agent, agent_kwargs_agent_only = create_agent(mdp_train, config_agent, config_env["unit_size"])

            agent_l.append(agent)
        
        mdp_train = mdp_train_l
        mdp_val = mdp_val_l
        mdp_test = None
        agent = agent_l

    else:
        if config_train["TRAINING_PARAMS"]["early_stopping_patience"] > 0 or config_train["TRAINING_PARAMS"]["save_best"]:
            print("creating validation and test set")
            mdp_train, mdp_val, mdp_test = create_env(config_env, create_val=True, actions_per_product=actions_per_product) 
        else:
            print("creating only test set")
            mdp_train, mdp_test = create_env(config_env, actions_per_product=actions_per_product)
            mdp_val = None

        create_agent = load_agent_function(config_train["agent"], config_agent["AGENTS"])
        agent, agent_kwargs_agent_only = create_agent(mdp_train, config_agent, config_env["unit_size"])

    return create_agent, mdp_train, mdp_val, mdp_test, agent, agent_kwargs_agent_only



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../../nbs/agents/rl_agents/algorithms/02_SAC_hybrid.ipynb.

# %% auto 0
__all__ = ['SAC_hybridPolicy', 'SAC_hybrid']

# %% ../../../../nbs/agents/rl_agents/algorithms/02_SAC_hybrid.ipynb 4
# General libraries
import numpy as np
from copy import deepcopy

# PyTorch
import torch
import torch.optim as optim
import torch.nn.functional as F

# Mushroom
from mushroom_rl.utils.parameters import Parameter
from mushroom_rl.policy import Policy
from mushroom_rl.algorithms.actor_critic import DeepAC
from mushroom_rl.utils.parameters import to_parameter
from mushroom_rl.utils.replay_memory import ReplayMemory
from mushroom_rl.approximators import Regressor
from mushroom_rl.approximators.parametric.torch_approximator import *
from mushroom_rl.utils.torch import to_float_tensor
from itertools import chain

# RL_OM
from ..sac_gumbel import GumbelSoftmax

# %% ../../../../nbs/agents/rl_agents/algorithms/02_SAC_hybrid.ipynb 6
class SAC_hybridPolicy(Policy):
    
    # TODO update docstring
    """
    Class used to implement the policy used by the Soft Actor-Critic
    algorithm. The policy is a Gaussian policy squashed by a tanh.
    This class implements the compute_action_and_log_prob and the
    compute_action_and_log_prob_t methods, that are fundamental for
    the internals calculations of the SAC algorithm.

    """
    def __init__(self, mu_approximator, sigma_approximator, discrete_approximator,
                 min_a, max_a, log_std_min, log_std_max, temperature=1.0):
        """
        Constructor.

        Args:
            mu_approximator (Regressor): a regressor computing mean in a given
                state;
            sigma_approximator (Regressor): a regressor computing the variance
                in a given state;
            discrete_approximator (Regressor): a regressor computing the discrete
                action disctribution in a given state;
            min_a (np.ndarray): a vector specifying the minimum action value
                for each component;
            max_a (np.ndarray): a vector specifying the maximum action value
                for each component.
            log_std_min ([float, Parameter]): min value for the policy log std;
            log_std_max ([float, Parameter]): max value for the policy log std;
            temperature ([float]): temperature for the Gumbel Softmax.

        """
        self._mu_approximator = mu_approximator
        self._sigma_approximator = sigma_approximator
        self._discrete_approximator = discrete_approximator
        
        self._temperature = torch.tensor(temperature)
        self._max_a = max_a[:mu_approximator.output_shape[0]]
        self._min_a = min_a[:mu_approximator.output_shape[0]]
        self._delta_a = to_float_tensor(.5 * (self._max_a - self._min_a), self.use_cuda)
        self._central_a = to_float_tensor(.5 * (self._max_a + self._min_a), self.use_cuda)

        self._log_std_min = to_parameter(log_std_min)
        self._log_std_max = to_parameter(log_std_max)

        self._eps_log_prob = 1e-6

        use_cuda = self._mu_approximator.model.use_cuda

        if use_cuda:
            self._delta_a = self._delta_a.cuda()
            self._central_a = self._central_a.cuda()

        self._add_save_attr(
            _mu_approximator='mushroom',
            _sigma_approximator='mushroom',
            _discrete_approximator='mushroom',
            _delta_a='torch',
            _central_a='torch',
            _log_std_min='mushroom',
            _log_std_max='mushroom',
            _eps_log_prob='primitive',
            _temperature='torch'
        )

    def __call__(self, state, action):
        raise NotImplementedError

    def draw_action(self, state):
        return self.compute_action_and_log_prob_t(
            state, compute_log_prob=False).detach().cpu().numpy()

    def draw_noisy_action(self, state):
        # Dummy call to draw_action
        return self.draw_action(state)

    def draw_action_mean_and_logits(self, state):
        # Continuous
        cont_mu_raw = self._mu_approximator.predict(state, output_tensor=True)
        a_cont = torch.tanh(cont_mu_raw)
        a_cont_true = a_cont * self._delta_a + self._central_a
        # Discrete
        # NOTE: Discrete approximator takes both state and continuous action as input (sequential policy)
        if isinstance(state, np.ndarray):
            if self._mu_approximator.model.use_cuda:
                state = torch.from_numpy(state).cuda()
            else:
                state = torch.from_numpy(state)
        logits = self._discrete_approximator.predict(torch.hstack((state, a_cont_true.detach())), output_tensor=True)

        return a_cont_true.detach().squeeze(), logits.detach().squeeze()
    
    def compute_action_and_log_prob(self, state):
        """
        Function that samples actions using the reparametrization trick and
        the log probability for such actions.

        Args:
            state (np.ndarray): the state in which the action is sampled.

        Returns:
            The actions sampled and the log probability as numpy arrays.

        """
        a, log_prob = self.compute_action_and_log_prob_t(state)
        return a.detach().cpu().numpy(), log_prob.detach().cpu().numpy()

    def compute_action_and_log_prob_t(self, state, compute_log_prob=True):
        """
        Function that samples actions using the reparametrization trick and,
        optionally, the log probability for such actions.

        Args:
            state (np.ndarray): the state in which the action is sampled;
            compute_log_prob (bool, True): whether to compute the log
            probability or not.

        Returns:
            The actions sampled and, optionally, the log probability as torch
            tensors.

        """
        # Continuous
        cont_dist = self.cont_distribution(state)
        a_cont_raw = cont_dist.rsample()
        a_cont = torch.tanh(a_cont_raw)
        a_cont_true = a_cont * self._delta_a + self._central_a

        # Discrete
        # NOTE: Discrete approximator takes both state and continuous action as input (sequential policy)
        discrete_dist = self.discrete_distribution(state, a_cont_true.detach()) # detach to avoid gradients of continuous through here
        a_discrete = discrete_dist.rsample()

        if compute_log_prob:
            # Continuous
            log_prob_cont = cont_dist.log_prob(a_cont_raw).sum(dim=1)
            log_prob_cont -= torch.log(1. - a_cont.pow(2) + self._eps_log_prob).sum(dim=1)
            # Discrete
            log_prob_discrete = discrete_dist.log_prob(a_discrete)
            return torch.hstack((a_cont_true, a_discrete)), log_prob_cont+log_prob_discrete
        else:
            return torch.hstack((a_cont_true, a_discrete))

    def cont_distribution(self, state):
        """
        Compute the continous (Gaussian) policy distribution in the given states.

        Args:
            state (np.ndarray): the set of states where the distribution is
                computed.

        Returns:
            The torch distribution for the provided states.

        """
        mu = self._mu_approximator.predict(state, output_tensor=True)
        log_sigma = self._sigma_approximator.predict(state, output_tensor=True)
        # Bound the log_std
        log_sigma = torch.clamp(log_sigma, self._log_std_min(), self._log_std_max())
        return torch.distributions.Normal(mu, log_sigma.exp())

    def discrete_distribution(self, state, a_cont):
        """
        Compute the discrete policy distribution (categorical) in the given states.

        Args:
            state (np.ndarray): the set of states where the distribution is
                computed.
            a_cont (torch tensor): the set of continuous actions, conditioned on 
                which, the discrete distribution is computed.

        Returns:
            The torch distribution for the provided states.

        """
        if isinstance(state, np.ndarray):
            if self._mu_approximator.model.use_cuda:
                state = torch.from_numpy(state).cuda()
            else:
                state = torch.from_numpy(state)

        logits = self._discrete_approximator.predict(torch.hstack((state, a_cont)), output_tensor=True)

        return GumbelSoftmax(temperature=self._temperature, logits=logits)

    
    def entropy(self, state=None):
        """
        Compute the entropy of the policy.

        Args:
            state (np.ndarray): the set of states to consider.

        Returns:
            The value of the entropy of the policy.

        """
        # Continuous dist and action
        cont_distr = self.cont_distribution(state)
        act_cont_raw = cont_distr.rsample()
        act_cont_true = torch.tanh(act_cont_raw) * self._delta_a + self._central_a

        # return sum of cont and discrete entropy
        return torch.mean(cont_distr.entropy()).detach().cpu().numpy().item() + torch.mean(self.discrete_distribution(state, act_cont_true).entropy()).detach().cpu().numpy().item()

    def reset(self):
        pass

    def set_weights(self, weights):
        """
        Setter.

        Args:
            weights (np.ndarray): the vector of the new weights to be used by
                the policy.

        """
        mu_weights = weights[:self._mu_approximator.weights_size]
        sigma_weights = weights[self._mu_approximator.weights_size:self._mu_approximator.weights_size+self._sigma_approximator.weights_size]
        discrete_weights = weights[self._mu_approximator.weights_size+self._sigma_approximator.weights_size:]

        self._mu_approximator.set_weights(mu_weights)
        self._sigma_approximator.set_weights(sigma_weights)
        self._discrete_approximator.set_weights(discrete_weights)

    def get_weights(self):
        """
        Getter.

        Returns:
             The current policy weights.

        """
        mu_weights = self._mu_approximator.get_weights()
        sigma_weights = self._sigma_approximator.get_weights()
        discrete_weights = self._discrete_approximator.get_weights()

        return np.concatenate([mu_weights, sigma_weights, discrete_weights])

    @property
    def use_cuda(self):
        """
        True if the policy is using cuda_tensors.
        """
        return self._mu_approximator.model.use_cuda

    def parameters(self):
        """
        Returns the trainable policy parameters, as expected by torch
        optimizers.

        Returns:
            List of parameters to be optimized.

        """
        return chain(self._mu_approximator.model.network.parameters(),
                     self._sigma_approximator.model.network.parameters(),
                     self._discrete_approximator.model.network.parameters())

# %% ../../../../nbs/agents/rl_agents/algorithms/02_SAC_hybrid.ipynb 8
class SAC_hybrid(DeepAC):
    """
    Soft Actor-Critic algorithm.
    "Soft Actor-Critic Algorithms and Applications".
    Haarnoja T. et al.. 2019.

    * With a Hybrid action space (A sequential discrete approximator takes as input the continous action and outputs the discrete part of the action)

    """
    def __init__(self, mdp_info, actor_mu_params, actor_sigma_params, actor_discrete_params,
                 actor_optimizer, critic_params, batch_size,
                 initial_replay_size, max_replay_size, warmup_transitions, tau,
                 lr_alpha, log_std_min=-20, log_std_max=2, temperature=1.0, target_entropy=None,
                 critic_fit_params=None):
        """
        Constructor.

        Args:
            actor_mu_params (dict): parameters of the actor mean approximator
                to build;
            actor_sigma_params (dict): parameters of the actor sigma
                approximator to build;
            actor_discrete_params (dict): parameters of the actor discrete distribution
                approximator to build;
            actor_optimizer (dict): parameters to specify the actor
                optimizer algorithm;
            critic_params (dict): parameters of the critic approximator to
                build;
            batch_size ((int, Parameter)): the number of samples in a batch;
            initial_replay_size (int): the number of samples to collect before
                starting the learning;
            max_replay_size (int): the maximum number of samples in the replay
                memory;
            warmup_transitions ([int, Parameter]): number of samples to accumulate in the
                replay memory to start the policy fitting;
            tau ([float, Parameter]): value of coefficient for soft updates;
            lr_alpha ([float, Parameter]): Learning rate for the entropy coefficient;
            log_std_min ([float, Parameter]): Min value for the policy log std;
            log_std_max ([float, Parameter]): Max value for the policy log std;
            temperature (float): the temperature for the softmax part of the gumbel reparametrization
            target_entropy (float, None): target entropy for the policy, if
                None a default value is computed ;
            critic_fit_params (dict, None): parameters of the fitting algorithm
                of the critic approximator.

        """
        self._critic_fit_params = dict() if critic_fit_params is None else critic_fit_params

        self._batch_size = to_parameter(batch_size)
        self._warmup_transitions = to_parameter(warmup_transitions)
        self._tau = to_parameter(tau)

        if target_entropy is None:
            self._target_entropy = -np.prod(mdp_info.action_space.shape).astype(np.float32)            
        else:
            self._target_entropy = target_entropy

        self._replay_memory = ReplayMemory(initial_replay_size, max_replay_size)

        if 'n_models' in critic_params.keys():
            assert critic_params['n_models'] == 2
        else:
            critic_params['n_models'] = 2

        target_critic_params = deepcopy(critic_params)

        self._critic_approximator = Regressor(TorchApproximator,
                                            **critic_params)
        self._target_critic_approximator = Regressor(TorchApproximator,
                                                    **target_critic_params)

        self._state_dim = actor_mu_params['input_shape'][0] # Store state dimensions for help in boosting (change in state spaces)
        actor_mu_approximator = Regressor(TorchApproximator,
                                          **actor_mu_params)
        actor_sigma_approximator = Regressor(TorchApproximator,
                                             **actor_sigma_params)
        actor_discrete_approximator = Regressor(TorchApproximator,
                                             **actor_discrete_params)
        self._actor_last_loss = None # Store actor loss for logging
        
        policy = SAC_hybridPolicy(actor_mu_approximator,
                           actor_sigma_approximator,
                           actor_discrete_approximator,
                           mdp_info.action_space.low,
                           mdp_info.action_space.high,
                           log_std_min,
                           log_std_max,
                           temperature)

        self._init_target(self._critic_approximator,
                          self._target_critic_approximator)

        self._log_alpha = torch.tensor(0., dtype=torch.float32)

        if policy.use_cuda:
            self._log_alpha = self._log_alpha.cuda().requires_grad_()
        else:
            self._log_alpha.requires_grad_()

        self._alpha_optim = optim.Adam([self._log_alpha], lr=lr_alpha)

        policy_parameters = chain(actor_mu_approximator.model.network.parameters(),
                                  actor_sigma_approximator.model.network.parameters(),
                                  actor_discrete_approximator.model.network.parameters())

        self._add_save_attr(
            _critic_fit_params='pickle',
            _batch_size='mushroom',
            _warmup_transitions='mushroom',
            _tau='mushroom',
            _target_entropy='primitive',
            _replay_memory='mushroom',
            _critic_approximator='mushroom',
            _target_critic_approximator='mushroom',
            _boosting='primitive',
            _state_dim='primitive',
            _log_alpha='torch',
            _alpha_optim='torch'
        )

        super().__init__(mdp_info, policy, actor_optimizer, policy_parameters)

    def fit(self, dataset):
        self._replay_memory.add(dataset)
        if self._replay_memory.initialized:
            state, action, reward, next_state, absorbing, _ = \
                self._replay_memory.get(self._batch_size())

            if self._replay_memory.size > self._warmup_transitions():
                action_new, log_prob = self.policy.compute_action_and_log_prob_t(state)
                loss = self._loss(state, action_new, log_prob)
                self._optimize_actor_parameters(loss)
                self._update_alpha(log_prob.detach())
                self._actor_last_loss = loss.detach().cpu().numpy() # Store actor loss for logging

            q_next = self._next_q(next_state, absorbing)
            q = reward + self.mdp_info.gamma * q_next
            
            self._critic_approximator.fit(state, action, q,
                                          **self._critic_fit_params)

            self._update_target(self._critic_approximator,
                                self._target_critic_approximator)

    def _loss(self, state, action_new, log_prob):
        q_0 = self._critic_approximator(state, action_new,
                                        output_tensor=True, idx=0)
        q_1 = self._critic_approximator(state, action_new,
                                        output_tensor=True, idx=1)

        q = torch.min(q_0, q_1)

        return (self._alpha * log_prob - q).mean()

    def _update_alpha(self, log_prob):
        alpha_loss = - (self._log_alpha * (log_prob + self._target_entropy)).mean()
        self._alpha_optim.zero_grad()
        alpha_loss.backward()
        self._alpha_optim.step()

    def _next_q(self, next_state, absorbing):
        """
        Args:
            next_state (np.ndarray): the states where next action has to be
                evaluated;
            absorbing (np.ndarray): the absorbing flag for the states in
                ``next_state``.

        Returns:
            Action-values returned by the critic for ``next_state`` and the
            action returned by the actor.

        """
        a, log_prob_next = self.policy.compute_action_and_log_prob(next_state)

        q = self._target_critic_approximator.predict(
            next_state, a, prediction='min') - self._alpha_np * log_prob_next
        q *= 1 - absorbing

        return q

    def _post_load(self):
        self._update_optimizer_parameters(self.policy.parameters())

    @property
    def _alpha(self):
        return self._log_alpha.exp()

    @property
    def _alpha_np(self):
        return self._alpha.detach().cpu().numpy()



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../../nbs/agents/rl_agents/pre_specified_agents/21_SAC_hybrid_gumbel.ipynb.

# %% auto 0
__all__ = ['SACHybridGumbel', 'SACHybridGumbelReversed', 'SACHybridGumbelSeparate']

# %% ../../../../nbs/agents/rl_agents/pre_specified_agents/21_SAC_hybrid_gumbel.ipynb 4
# General libraries
import numpy as np

# Torch
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

# Networks
from ...networks.critics import CriticNetworkStateAction
from ...networks.actors import ActorNetwork

# Algorithms
from ..sac_hybrid import SAC_hybrid
from ..sac_hybrid_separate import SAC_hybrid_separate
from ..sac_hybrid_reversed import SAC_hybrid_reversed

from mushroom_rl.core.serialization import Serializable

# Processors
from ...processors.processors import  OneHotHybridtoContinuous


# %% ../../../../nbs/agents/rl_agents/pre_specified_agents/21_SAC_hybrid_gumbel.ipynb 5
class SACHybridGumbel():

    """
    Soft Actor Critic (SAC) agent with hybrid Gaussian-Gumbel policy.
    ! Currently works only for binary discrete actions.

    Args:
        mdp_info (MDPInfo): Contains relevant information about the environment.
        learning_rate_actor (float): Learning rate for the actor.
        learning_rate_critic (float): Learning rate for the critic.
        learning_rate_alpha (float): Learning rate for the temperature parameter.
        initial_replay_size (int): Number of transitions to save in the replay buffer during startup.
        max_replay_size (int): Maximum number of transitions to save in the replay buffer.
        batch_size (int): Number of transitions to sample each time experience is replayed.
        n_features (int): Number of features for the hidden layers of the networks.
        n_features_discrete (int): Number of features for the hidden layers of the discrete network. If set to None will use the same as n_features.
        warmup_transitions (int): Number of transitions to replay before starting to update the networks.
        lr_alpha (float): Learning rate for the temperature parameter.
        tau (float): Parameter for the soft update of the target networks.
        optimizer (torch.optim): Optimizer to use for the networks.
        squeeze_output (bool): Whether to squeeze the output of the actor network or not.
        use_cuda (bool): Whether to use CUDA or not. If True and not available, it will use CPU.
        agent_name (str): Name of the agent. If set to None will use some default name.

    """

    def __init__(
            self,
            mdp_info,
            learning_rate_actor = 3e-4,
            learning_rate_critic = None,
            learning_rate_alpha = 3e-4,
            initial_replay_size = 64,
            max_replay_size = 50000,
            batch_size = 64,
            n_features = 64,
            n_features_discrete = None,
            warmup_transitions = 100,
            lr_alpha = 3e-4,
            tau = 0.005,
            target_entropy = None,
            optimizer = optim.Adam,
            squeeze_output = True,
            use_cuda = True,
            agent_name = None):
        
        use_cuda = use_cuda and torch.cuda.is_available()

        actor_input_shape = mdp_info.observation_space.shape
        actor_output_shape = (mdp_info.action_space.shape[0],)
        actor_discrete_input_shape = (mdp_info.observation_space.shape[0] + mdp_info.action_space.shape[0],)
        actor_discrete_output_shape = (2,) # Currently hardcoded for binary discrete actions

        if learning_rate_critic is None:
            learning_rate_critic = learning_rate_actor


        actor_mu_params = dict(network=ActorNetwork,
                                n_features=n_features,
                                input_shape=actor_input_shape,
                                output_shape=actor_output_shape,
                                use_cuda=use_cuda)

        actor_sigma_params = dict(network=ActorNetwork,
                                    n_features=n_features,
                                    input_shape=actor_input_shape,
                                    output_shape=actor_output_shape,
                                    use_cuda=use_cuda)
        
        actor_discrete_params = dict(network=ActorNetwork,
                                    n_features=n_features if n_features_discrete is None else n_features_discrete,
                                    input_shape=actor_discrete_input_shape,
                                    output_shape=actor_discrete_output_shape,
                                    use_cuda=use_cuda)
        
        actor_optimizer = {'class': optimizer,
                    'params': {'lr': learning_rate_alpha}} 
        
        critic_input_shape = (actor_input_shape[0] + actor_output_shape[0]*3,)
        critic_params = dict(network=CriticNetworkStateAction,
                        optimizer={'class': optim.Adam,
                                'params': {'lr': learning_rate_critic}}, 
                        loss=F.mse_loss,
                        n_features=n_features,
                        input_shape=critic_input_shape,
                        output_shape=(1,),
                        squeeze_output=squeeze_output,
                        use_cuda=use_cuda)

        self.agent = SAC_hybrid(mdp_info, actor_mu_params, actor_sigma_params,
                actor_discrete_params, actor_optimizer, critic_params, batch_size, initial_replay_size,
                max_replay_size, warmup_transitions, tau, lr_alpha,
                critic_fit_params=None)
        
        if agent_name is None:
            self.agent.name = 'SAC_hybrid_gumbel'
        else:
            self.agent.name = agent_name

        self.agent.add_postprocessor(OneHotHybridtoContinuous(mdp_info.action_space.shape[0]))

    def __getattr__(self, attr):
        return getattr(self.agent, attr)



# %% ../../../../nbs/agents/rl_agents/pre_specified_agents/21_SAC_hybrid_gumbel.ipynb 6
class SACHybridGumbelReversed():

    """
    Soft Actor Critic (SAC) agent with hybrid Gaussian-Gumbel policy. The order of the actor
    Networks is reversed in comparison to the original implementation (SACHybridGumbel)
    ! Currently works only for binary discrete actions.

   Args:
        mdp_info (MDPInfo): Contains relevant information about the environment.
        learning_rate_actor (float): Learning rate for the actor.
        learning_rate_critic (float): Learning rate for the critic.
        learning_rate_alpha (float): Learning rate for the temperature parameter.
        initial_replay_size (int): Number of transitions to save in the replay buffer during startup.
        max_replay_size (int): Maximum number of transitions to save in the replay buffer.
        batch_size (int): Number of transitions to sample each time experience is replayed.
        n_features (int): Number of features for the hidden layers of the networks.
        n_features_discrete (int): Number of features for the hidden layers of the discrete network. If set to None will use the same as n_features.
        warmup_transitions (int): Number of transitions to replay before starting to update the networks.
        lr_alpha (float): Learning rate for the temperature parameter.
        tau (float): Parameter for the soft update of the target networks.
        optimizer (torch.optim): Optimizer to use for the networks.
        use_cuda (bool): Whether to use CUDA or not. If True and not available, it will use CPU.
        agent_name (str): Name of the agent. If set to None will use some default name.

    """

    def __init__(
            self,
            mdp_info,
            learning_rate_actor = 3e-4,
            learning_rate_critic = None,
            learning_rate_alpha = 3e-4,
            initial_replay_size = 64,
            max_replay_size = 50000,
            batch_size = 64,
            n_features = 64,
            n_features_discrete = None,
            warmup_transitions = 100,
            lr_alpha = 3e-4,
            tau = 0.005,
            target_entropy = None,
            optimizer = optim.Adam,
            squeeze_output = True,
            use_cuda = True,
            agent_name = None):
        
        use_cuda = use_cuda and torch.cuda.is_available()

        actor_input_shape = mdp_info.observation_space.shape
        actor_output_shape = (mdp_info.action_space.shape[0],)
        actor_continuous_input_shape = (actor_input_shape[0]+2,)
        actor_discrete_output_shape = (2,) # Currently hardcoded for binary discrete actions

        if learning_rate_critic is None:
            learning_rate_critic = learning_rate_actor

        # Changer here: Now there is an additional input representing the taken discrete action
        actor_mu_params = dict(network=ActorNetwork,
                                n_features=n_features,
                                input_shape=actor_continuous_input_shape , # Currently hardcoded for binary discrete actions
                                output_shape=actor_output_shape,
                                use_cuda=use_cuda)

        # Changer here: Now there is an additional input representing the taken discrete action
        actor_sigma_params = dict(network=ActorNetwork,
                                    n_features=n_features,
                                    input_shape=actor_continuous_input_shape , # Currently hardcoded for binary discrete actions
                                    output_shape=actor_output_shape,
                                    use_cuda=use_cuda)
        
        # Change here: Only state as input for discrete network. Same as for sepearate agent
        actor_discrete_params = dict(network=ActorNetwork,
                            n_features = n_features if n_features_discrete is None else n_features_discrete,
                            input_shape=actor_input_shape,
                            output_shape=actor_discrete_output_shape,
                            use_cuda=use_cuda)
        
        actor_optimizer = {'class': optimizer,
                    'params': {'lr': learning_rate_alpha}} 
        
        critic_input_shape = (actor_input_shape[0] + actor_output_shape[0]*3,)
        critic_params = dict(network=CriticNetworkStateAction,
                        optimizer={'class': optim.Adam,
                                'params': {'lr': learning_rate_critic}}, 
                        loss=F.mse_loss,
                        n_features=n_features,
                        input_shape=critic_input_shape,
                        output_shape=(1,),
                        squeeze_output=squeeze_output,
                        use_cuda=use_cuda)

        # Change here: Use reversed SAC algorithm
        self.agent = SAC_hybrid_reversed(mdp_info, actor_mu_params, actor_sigma_params,
                actor_discrete_params, actor_optimizer, critic_params, batch_size, initial_replay_size,
                max_replay_size, warmup_transitions, tau, lr_alpha,
                critic_fit_params=None)
        
        # Change here: Different default name
        if agent_name is None:
            self.agent.name = 'SAC_hybrid_gumbel_reversed'
        else:
            self.agent.name = agent_name

        self.agent.add_postprocessor(OneHotHybridtoContinuous(mdp_info.action_space.shape[0]))

    def __getattr__(self, attr):
        return getattr(self.agent, attr)

# %% ../../../../nbs/agents/rl_agents/pre_specified_agents/21_SAC_hybrid_gumbel.ipynb 7
class SACHybridGumbelSeparate():

    """
    Soft Actor Critic (SAC) agent with hybrid Gaussian-Gumbel policy. The contiuous and discrete
    Actor networks are independent, different from the original implementation (SACHybridGumbel)
    ! Currently works only for binary discrete actions.

   Args:
        mdp_info (MDPInfo): Contains relevant information about the environment.
        learning_rate_actor (float): Learning rate for the actor.
        learning_rate_critic (float): Learning rate for the critic.
        learning_rate_alpha (float): Learning rate for the temperature parameter.
        initial_replay_size (int): Number of transitions to save in the replay buffer during startup.
        max_replay_size (int): Maximum number of transitions to save in the replay buffer.
        batch_size (int): Number of transitions to sample each time experience is replayed.
        n_features (int): Number of features for the hidden layers of the networks.
        n_features_discrete (int): Number of features for the hidden layers of the discrete network. If set to None will use the same as n_features.
        warmup_transitions (int): Number of transitions to replay before starting to update the networks.
        lr_alpha (float): Learning rate for the temperature parameter.
        tau (float): Parameter for the soft update of the target networks.
        optimizer (torch.optim): Optimizer to use for the networks.
        use_cuda (bool): Whether to use CUDA or not. If True and not available, it will use CPU.
        agent_name (str): Name of the agent. If set to None will use some default name.

    """

    def __init__(
            self,
            mdp_info,
            learning_rate_actor = 3e-4,
            learning_rate_critic = None,
            learning_rate_alpha = 3e-4,
            initial_replay_size = 64,
            max_replay_size = 50000,
            batch_size = 64,
            n_features = 64,
            n_features_discrete = None,
            warmup_transitions = 100,
            lr_alpha = 3e-4,
            tau = 0.005,
            target_entropy = None,
            optimizer = optim.Adam,
            squeeze_output = True,
            use_cuda = True,
            agent_name = None):
        
        use_cuda = use_cuda and torch.cuda.is_available()

        actor_input_shape = mdp_info.observation_space.shape
        actor_output_shape = (mdp_info.action_space.shape[0],)
        actor_discrete_output_shape = (2,) # Currently hardcoded for binary discrete actions

        if learning_rate_critic is None:
            learning_rate_critic = learning_rate_actor

        actor_mu_params = dict(network=ActorNetwork,
                                n_features=n_features,
                                input_shape=actor_input_shape,
                                output_shape=actor_output_shape,
                                use_cuda=use_cuda)

        actor_sigma_params = dict(network=ActorNetwork,
                                    n_features=n_features,
                                    input_shape=actor_input_shape,
                                    output_shape=actor_output_shape,
                                    use_cuda=use_cuda)
        
        # Change here: Only state as input for discrete network. Same as for reversed agent
        actor_discrete_params = dict(network=ActorNetwork,
                            n_features=n_features if n_features_discrete is None else n_features_discrete,
                            input_shape=actor_input_shape,
                            output_shape=actor_discrete_output_shape,
                            use_cuda=use_cuda)
        
        actor_optimizer = {'class': optimizer,
                    'params': {'lr': learning_rate_alpha}} 
        
        critic_input_shape = (actor_input_shape[0] + actor_output_shape[0]*3,) # Currently hardcoded for binary discrete actions
        critic_params = dict(network=CriticNetworkStateAction,
                        optimizer={'class': optim.Adam,
                                'params': {'lr': learning_rate_critic}}, 
                        loss=F.mse_loss,
                        n_features=n_features,
                        input_shape=critic_input_shape,
                        output_shape=(1,),
                        squeeze_output=squeeze_output,
                        use_cuda=use_cuda)

        # Change here: Use separate SAC algorithm
        self.agent = SAC_hybrid_separate(mdp_info, actor_mu_params, actor_sigma_params,
                actor_discrete_params, actor_optimizer, critic_params, batch_size, initial_replay_size,
                max_replay_size, warmup_transitions, tau, lr_alpha,
                critic_fit_params=None)
        
        # Change here: Different default name
        if agent_name is None:
            self.agent.name = 'SAC_hybrid_gumbel_separate'
        else:
            self.agent.name = agent_name

        self.agent.add_postprocessor(OneHotHybridtoContinuous(mdp_info.action_space.shape[0]))

    def __getattr__(self, attr):
        return getattr(self.agent, attr)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/utils/02_ML_Tools.ipynb.

# %% auto 0
__all__ = ['LRSchedulerPerStep']

# %% ../nbs/utils/02_ML_Tools.ipynb 3
# General libraries:
# import numpy as np

import wandb
import datetime
import torch


# %% ../nbs/utils/02_ML_Tools.ipynb 4
class LRSchedulerPerStep():
    """
    Learning rate scheduler from Attention is all you need paper
    Change: added basic LR as parameter rather than automated based on model dimension

    """
    
    def __init__(self, optimizer, base_learning_rate=0.0001, warmup=4000):

        # Ensure optimizer is a PyTorch optimizer
        if not isinstance(optimizer, torch.optim.Optimizer):
            raise ValueError('Optimizer must be a PyTorch optimizer')
        
        self.optimizer = optimizer
        self.basic = base_learning_rate
        self.warm = warmup**-1.5
        self.step_num = 0
        self.scaling_factor = 1/warmup**-0.5 # ensures that the peak realtive to the base lr is always 1

        print("set up schedule")
        
    def step(self):
        self.step_num += 1
        lr = self.basic * self.scaling_factor * min(self.step_num**-0.5, self.step_num*self.warm)
        
        for param_group in self.optimizer.param_groups:
            param_group['lr'] = lr
		



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/environments/01_multi_period_inventory.ipynb.

# %% auto 0
__all__ = ['MultiPeriodEnv', 'MultiPeriodEnvMP']

# %% ../../nbs/environments/01_multi_period_inventory.ipynb 4
# General libraries:
import numpy as np
    
# build env:
from mushroom_rl.core import Environment, MDPInfo
from mushroom_rl.utils.spaces import Discrete, Box
from .feature_converters import get_time_series_features

# others
from abc import ABC, abstractmethod

# %% ../../nbs/environments/01_multi_period_inventory.ipynb 6
class MultiPeriodEnv(Environment, ABC):
    def __init__(self,
                 
                ### demand data
                demand,

                ### feature data
                features = None,

                mask = None, # optional mask to mask out certain actions (e.g., if product not offered), shape(timesteps, num_products)
                feature_map = None, # optional feature map to map features to products, shape(num_features, num_products)

                ### cost related to overage and underate
                overage_cost = np.array([0]), 
                underage_cost = np.array([0]), #! check if there can be a case where this is different from penalty cost

                ### cost related to ordering
                fixed_ordering_cost = np.array([0]), 
                variable_ordering_cost = np.array([0]), 

                ### demand parameters
                max_demand_backlog = np.array([0]), 

                ### inventory parameters
                inventory_allowed = True,
                start_inventory = np.array([0]), 
                inventory_cap = np.array([np.inf]), 
                
                ### cost related to inventory
                holding_cost = None, 

                ### ordering parameters
                lead_time = np.array([0]), 
                lead_time_stochastic = None, # other values: "gamma".
                lead_time_variance = np.array([0]), # not implemented, only relevant for stochastic lead times with variance as factor, currently only gamma distribution is implemented
                max_lead_time = None, # factor dependent on mean demand
                actions_per_product = 1,
                order_limit_low = None, 
                order_limit_high = None, 

                ### other parameters
                horizon = 365, # lenght of each episode in simulation
                gamma = 1, # discount factor
                observation_space_kwargs = None,

                ### logic of step-method
                calculations = None,

                ### Precision of state
                precision = 5,

                lag_window = 0,
                dynamic_input = None, # Not yet implemented for this kind of enve

                normalize_env_state = False,
                normalize_reward = False,

                record_state = False
                ):
        """
        Multi-period inventory environment covering all major single-echelon inventory models with decision on quantity.

        Note: Except True/False parameters and horizon/gamma, all parameters are arrays to set the parameter for each product individually.

        args:
            demand: historical observations of demand
            features: features related to demand - normalize features before passing to environment
            overage_cost: overage cost incurred per unit in each period
            underage_cost: underage cost incurred per unit in each period
            fixed_ordering_cost: fixed cost incurred per order
            variable_ordering_cost: variable cost incurred per unit ordered (note: related to ordering only, not cogs)
            max_demand_backlog: How much demand can be backlogged. If set to zero: Lost Sales model
            inventory_allowed: whether inventory is allowed. False: Excess orders will be thrown away
            start_inventory: initial inventory
            inventory_cap: maximum inventory allowed
            holding_cost: holding cost incurred per unit in each period
            lead_time: lead time between ordering and receiving the order
            lead_time_variance: variance in lead time
            order_limit_low: maximum order allowed
            order_limit_high: minimum order allowed
            horizon: lenght of each episode in simulation
            gamma: discount factor
            observation_space_kwargs: kwargs for observation space
            calculations: function that calculates the next state and reward in the step function
            precision: precision of state variables (decimal places)
        """

        self.dynamic_input = dynamic_input
        if dynamic_input is not None:
            raise NotImplementedError("Dynamic input not yet implemented for this kind of environment")

        self.demand = demand
        self.avg_demand = np.mean(demand, axis = 0)

        self.lag_window = lag_window

        self.num_observations = demand.shape[0] # number of observations
        self.full_demand = False # demand without initial lag window
        self.num_products = demand.shape[1] # number of products
        self.features = features
        if self.features is not None:
            self.num_features = features.shape[1] # number of features
        else:   
            self.num_features = 0
        
        self.mask=mask

        self.feature_map = feature_map

        # print(demand.shape)
        # print(features.shape)
        # print("num_obs:", self.num_observations)
        # print("num_prod:", self.num_products)
        # print("num_feat:", self.num_features)

        self.overage_cost = self.convert_to_numpy_array(overage_cost, expand = True)
        self.underage_cost = self.convert_to_numpy_array(underage_cost, expand = True)

        # print("overage_cost:", self.overage_cost)
        # print("underage_cost:", self.underage_cost)


        self.fixed_ordering_cost = self.convert_to_numpy_array(fixed_ordering_cost, expand = True)
        self.variable_ordering_cost = self.convert_to_numpy_array(variable_ordering_cost, expand = True)

        self.max_demand_backlog = self.convert_to_numpy_array(max_demand_backlog, expand = True)

        # print("max_demand_backlog:", self.max_demand_backlog)
        # print("variable_ordering_cost:", self.variable_ordering_cost)
        # print("fixed_ordering_cost:", self.fixed_ordering_cost)
        

        self.inventory_allowed = inventory_allowed

        if not self.inventory_allowed:
            assert np.all(start_inventory == 0), "If inventory is not allowed, start_inventory must be 0"
        self.start_inventory = self.convert_to_numpy_array(start_inventory, expand = True)
        # print("setting inventory cap")
        self.inventory_cap = self.convert_to_numpy_array(inventory_cap, expand = True)

        if normalize_env_state:
            assert np.all(inventory_cap != np.inf), "If inventory is normalized, inventory_cap must be a finite number"
        
        # print("start_inventory:", self.start_inventory)
        # print("inventory_cap:", self.inventory_cap)

        # print("set inventory cap")

        if not self.inventory_allowed:
            assert holding_cost is None, "If inventory is not allowed, holding_cost must be None"
        # print("setting holding cost")
        self.holding_cost = self.convert_to_numpy_array(holding_cost, expand = True)

        # print("holding_cost:", self.holding_cost)
        # print("staring setting LT")

        if np.all(lead_time == 0):
            self.use_order_pipeline = False
        else:
            self.use_order_pipeline = True

        self.lead_time = self.convert_to_numpy_array(lead_time, expand = True)
        self.lead_time_stochastic = lead_time_stochastic
        self.lead_time_variance = self.convert_to_numpy_array(lead_time_variance, expand = True)

        # print("lead_time:", self.lead_time)
        # print("lead_time_variance:", self.lead_time_variance)

        if lead_time_stochastic is None:
            assert np.all(lead_time_variance == 0), "If lead_time_stochastic is None, lead_time_variance must be 0"
            
            self.max_lead_time = np.max(self.lead_time)
            self.max_observation_lead_time = np.max(self.lead_time)
            # print("set LT")

        else:
            if lead_time_stochastic == "gamma":
                if np.any(lead_time_variance == 1):
                    raise ValueError("Variance of lead time must be greater than 1 for gamma distribution, otherwise it is deterministically 1")
            self.max_lead_time = int(np.ceil(max_lead_time * np.max(self.lead_time))) # max lead time is factor of mean
            self.max_observation_lead_time = self.lag_window # the agnet can look back according to the lag window

        if order_limit_low is None:
            self.order_limit_low = np.zeros(self.num_products)
            self.order_limit_high = np.ones(self.num_products)
        
        self.order_limit_low = self.convert_to_numpy_array(order_limit_low, expand = True)
        self.order_limit_high = self.convert_to_numpy_array(order_limit_high, expand = True)

        # print("order_limit_low:", self.order_limit_low.shape)
        # print("order_limit_high:", self.order_limit_high.shape)

        order_limit_low = np.repeat(self.order_limit_low, actions_per_product)
        order_limit_high = np.repeat(self.order_limit_high, actions_per_product)

        self.action_space = Box(order_limit_low, order_limit_high, shape=(self.num_products*actions_per_product,))
        # print("action space:", self.action_space.shape)


        # Set observation space
        if observation_space_kwargs is None:
            observation_space_kwargs = {}

        # print("obs kwargs:", observation_space_kwargs)
        self.set_observation_space(**observation_space_kwargs)

        self.calculations = calculations

        # Set MDP info
        mdp_info = MDPInfo(observation_space=self.observation_space,
                           action_space=self.action_space,
                           horizon=horizon,
                           gamma=gamma)
        
        super().__init__(mdp_info)

        self.precision = precision

        self.normalize_reward = normalize_reward

        # print("normalioze rewrad in env class:", self.normalize_reward)
        self.normalize_env_state = normalize_env_state

        # print("done, now resetting")

        self.record_state = record_state

        # Set state variables
        self.reset()

    def set_observation_space(self,
                              features = False,
                              inventory = False,
                              order_pipeline = False,
                              demand_backlog = False
                              ):
        """
        Set observation space. This is used to customize the observation space.

        args:
            features: whether to include features in the observation space
            inventory: whether to include inventory in the observation space
            order_pipeline: whether to include order pipeline in the observation space
            demand_backlog: whether to include demand backlog in the observation space
        """

        features_space = self.num_features if features else 0
        inventory_space = self.num_products if inventory else 0
        order_pipeline_space = np.sum(self.num_products * np.max(self.max_observation_lead_time)) if order_pipeline else 0
        demand_backlog_space = self.num_products if demand_backlog else 0
        total_obs_space_length = features_space + inventory_space + order_pipeline_space + demand_backlog_space

        assert total_obs_space_length > 0, "Need at least one observation dimension"

        low_limit = np.zeros(total_obs_space_length)
        high_limit = np.zeros(total_obs_space_length)

        if features:
            low_limit[:self.num_features] = -np.ones(self.num_features)
            high_limit[:self.num_features] = np.ones(self.num_features)
        else:
            self.num_features = 0

        if order_pipeline:
            slots_occupied = 0
            for product in range(self.num_products):
                # Note: the length of the order pipeline can vary by product. Note that for RNN processing we need to pad the order pipeline to the same length in the preprocessing step
                low_limit[self.num_features+slots_occupied:self.num_features+slots_occupied+self.max_observation_lead_time] = self.order_limit_low[product]
                high_limit[self.num_features+slots_occupied:self.num_features+slots_occupied+self.max_observation_lead_time] = self.order_limit_high[product]
                slots_occupied += self.max_observation_lead_time
            order_pipeline_space = slots_occupied  # Keep track of the total order pipeline space used
        
        if inventory:
            # Update indices to reflect order_pipeline's position before inventory
            low_limit[self.num_features+order_pipeline_space:self.num_features+order_pipeline_space+inventory_space] = np.zeros(inventory_space) # assuming the low inventory end is always 0
            high_limit[self.num_features+order_pipeline_space:self.num_features+order_pipeline_space+inventory_space] = self.inventory_cap

        if demand_backlog:
            # Update indices to include both order_pipeline_space and inventory_space before demand_backlog
            low_limit[self.num_features+order_pipeline_space+inventory_space:self.num_features+order_pipeline_space+inventory_space+demand_backlog_space] = 0
            high_limit[self.num_features+order_pipeline_space+inventory_space:self.num_features+order_pipeline_space+inventory_space+demand_backlog_space] = self.max_demand_backlog

        # print("features space:", features_space)
        # print("inventory space:", inventory_space)
        # print("order pipeline space:", order_pipeline_space)
        # print("demand backlog space:", demand_backlog_space)
        # print("length:", total_obs_space_length)
        self.observation_space = Box(low_limit, high_limit, shape=(total_obs_space_length,))

    def reset(self, state = None):

        # TODO: Check in mushroom where state would come from and implement here
        if self.num_observations == self._mdp_info.horizon:
            self.period = 0
        else:
            self.period = np.random.choice(self.num_observations-self._mdp_info.horizon)
        #print("reset with period {}".format(self.period))
        self.demand_backlog = np.zeros(self.num_products)
        self.inventory = self.start_inventory.copy()
        self.order_pipeline = np.zeros((self.num_products,self.max_observation_lead_time))
        self.lead_time_realizations = np.zeros((self.num_products, self.max_lead_time))
        self.full_order_pipeline = np.zeros((self.num_products, self.max_lead_time))

        self.set_observation_state()

        return self.observation_state

    def set_observation_state(self):
        
        # TODO: check if this is computationally expensive
        # TODO: make sure this is consistent if some elements are there but not tracked in state

        if self.period >= self.num_observations:
            self.observation_state = np.zeros(self.observation_space.shape)
        
        else:
            
            self.observation_state = np.zeros(self.observation_space.shape)

            if self.num_features > 0:
                self.observation_state[:self.num_features] = self.features[self.period]

            if self.inventory_allowed:

                if self.normalize_env_state:
                    inv = self.inventory/self.inventory_cap
                else:
                    inv = self.inventory

                self.observation_state[self.num_features:self.num_features+self.num_products] = inv

            if self.use_order_pipeline:
                slots_occupied = 0
                for product in range(self.num_products):

                    # print("pipeline_raw:", self.order_pipeline[product])
                    
                    if self.normalize_env_state:
                        prod = self.order_pipeline[product]/self.order_limit_high[product]
                    else:
                        prod = self.order_pipeline[product]

                    # print(prod)

                    self.observation_state[self.num_features+self.num_products+slots_occupied:self.num_features+self.num_products+slots_occupied+self.max_observation_lead_time] = prod
                    slots_occupied += self.max_observation_lead_time

            if np.max(self.max_demand_backlog) > 0:
                self.observation_state[self.num_features+self.num_products+np.sum(self.num_products * self.lead_time):] = self.demand_backlog

            self.observation_state = np.round(self.observation_state, self.precision)

            # print("inventory:", np.round(self.inventory, 2))
            # print("order pipeline:", np.round(self.order_pipeline, 2))
            # print("_______")


    def step(self, action):

        absorbing = False

        reward, info = self.calculations(self, action)

        self.set_observation_state()
        
        # if self.normalize_reward:
        #     print(self.observation_state)

        return self.observation_state, reward, absorbing, info
    
    def render(self):
        pass

    def convert_to_numpy_array(self, value, expand = False):
        
        """
        Convert a scalar, list, or numpy array to a numpy array.
        
        Parameters:
        - value (scalar, list, numpy.ndarray): The input value.
        
        Returns:
        - numpy.ndarray: The converted numpy array.
        """

        if np.isscalar(value):
            value = np.array([value])
        elif isinstance(value, list):
            value = np.array(value)
        elif isinstance(value, np.ndarray):
            value = value
        elif value is None:
            value = None
        else:
            raise ValueError("Invalid input type. Expected scalar, list, or numpy array.")

        if expand:
            if value is not None:
                if value.size == 1:
                    value = np.repeat(value, self.num_products)
                elif value.size == self.num_products:
                    value = value
                else:
                    raise ValueError("Invalid input size. Expected scalar, list, or numpy array of size 1 or num_products.")
        return value

# %% ../../nbs/environments/01_multi_period_inventory.ipynb 7
class MultiPeriodEnvMP(Environment, ABC):
    def __init__(self,
                 
                ### demand data
                demand,

                ### feature data
                features = None,

                features_timeless = None,

                mask = None, # optional mask to mask out certain actions (e.g., if product not offered), shape(timesteps, num_products)
                feature_map = None, # optional feature map to map features to products, shape(num_features, num_products)

                ### cost related to overage and underate
                overage_cost = np.array([0]), 
                underage_cost = np.array([0]), #! check if there can be a case where this is different from penalty cost

                ### cost related to ordering
                fixed_ordering_cost = np.array([0]), 
                variable_ordering_cost = np.array([0]), 

                ### demand parameters
                max_demand_backlog = np.array([0]), 

                ### inventory parameters
                inventory_allowed = True,
                start_inventory = np.array([0]), 
                inventory_cap = np.array([np.inf]), 
                
                ### cost related to inventory
                holding_cost = None, 

                ### ordering parameters
                lead_time = np.array([0]), 
                lead_time_variance = np.array([0]), 
                max_lead_time = None,
                actions_per_product = 1,
                order_limit_low = None, 
                order_limit_high = None, 

                ### other parameters
                horizon = 365, # lenght of each episode in simulation
                gamma = 1, # discount factor
                observation_space_kwargs = None,

                ### logic of step-method
                calculations = None,

                ### Precision of state
                precision = 5,

                lag_window = 0,
                dynamic_input = None, 

                normalize_env_state = False,
                normalize_reward = False,

                record_state = False,

                additional_lag_features = None,
                include_alL_additional_lag_features = True
                ):
        """
        Multi-period inventory environment covering all major single-echelon inventory models with decision on quantity.

        Note: Except True/False parameters and horizon/gamma, all parameters are arrays to set the parameter for each product individually.

        args:
            demand: historical observations of demand
            features: features related to demand - normalize features before passing to environment
            overage_cost: overage cost incurred per unit in each period
            underage_cost: underage cost incurred per unit in each period
            fixed_ordering_cost: fixed cost incurred per order
            variable_ordering_cost: variable cost incurred per unit ordered (note: related to ordering only, not cogs)
            max_demand_backlog: How much demand can be backlogged. If set to zero: Lost Sales model
            inventory_allowed: whether inventory is allowed. False: Excess orders will be thrown away
            start_inventory: initial inventory
            inventory_cap: maximum inventory allowed
            holding_cost: holding cost incurred per unit in each period
            lead_time: lead time between ordering and receiving the order
            lead_time_variance: variance in lead time
            order_limit_low: maximum order allowed
            order_limit_high: minimum order allowed
            horizon: lenght of each episode in simulation
            gamma: discount factor
            observation_space_kwargs: kwargs for observation space
            calculations: function that calculates the next state and reward in the step function
            precision: precision of state variables (decimal places)
        """

        self.lag_window = lag_window
        self.dynamic_input = dynamic_input

        if additional_lag_features is not None:
            if include_alL_additional_lag_features:
                self.full_additional_context = lag_window + np.max(additional_lag_features)
            else:
                self.full_additional_context = lag_window # where not features, they will be filled with zeros
        else:
            self.full_additional_context = lag_window

        if dynamic_input is not None:
            self.update_engineered_features = True
            if self.dynamic_input < lag_window:
                demand = demand[lag_window-self.dynamic_input:]
                self.num_observations = demand.shape[0]-self.dynamic_input
            else:
                self.num_observations = demand.shape[0]-lag_window
        else:
            self.num_observations = demand.shape[0]-self.full_additional_context  # number of observations # in this environment, the demand will incldue lag-window observations in the beginning
            self.update_engineered_features = False

        self.demand = demand
        self.avg_demand = np.mean(demand, axis = 0)

        self.full_demand = True # demand with initial lag window
        
        self.num_products = demand.shape[1] # number of products
        self.features = [features, features_timeless] 

        if features is not None:

            if additional_lag_features is not None:
                # insert on 1 at the beginnig of the lst
                lag_features = [1] + additional_lag_features 
            else:
                lag_features = [1]
            self.lag_features = np.array(lag_features)

            if dynamic_input is not None:
                self.features[0] = self.features[0][lag_window-self.dynamic_input:]
            
            # number of overarching features (some may be partally product specific)
            self.num_features = feature_map.shape[1]
            self.num_features += 2 # price and sold information. # TODO: Automate this instead of hardcoding
            self.num_features += len(lag_features) # lag demand of previous step

            if features_timeless is not None:
                self.num_features += features_timeless.shape[1]

        else:   
            self.num_features = 0
        
        self.mask=mask

        self.feature_map = feature_map

        # print(demand.shape)
        # print(features.shape)
        # print("num_obs:", self.num_observations)
        # print("num_prod:", self.num_products)
        # print("num_feat:", self.num_features)

        self.overage_cost = self.convert_to_numpy_array(overage_cost, expand = True)
        self.underage_cost = self.convert_to_numpy_array(underage_cost, expand = True)

        # print("overage_cost:", self.overage_cost)
        # print("underage_cost:", self.underage_cost)

        self.fixed_ordering_cost = self.convert_to_numpy_array(fixed_ordering_cost, expand = True)
        self.variable_ordering_cost = self.convert_to_numpy_array(variable_ordering_cost, expand = True)

        self.max_demand_backlog = self.convert_to_numpy_array(max_demand_backlog, expand = True)

        # print("max_demand_backlog:", self.max_demand_backlog)
        # print("variable_ordering_cost:", self.variable_ordering_cost)
        # print("fixed_ordering_cost:", self.fixed_ordering_cost)
        

        self.inventory_allowed = inventory_allowed

        if not self.inventory_allowed:
            assert np.all(start_inventory == 0), "If inventory is not allowed, start_inventory must be 0"
        self.start_inventory = self.convert_to_numpy_array(start_inventory, expand = True)
        # print("setting inventory cap")
        self.inventory_cap = self.convert_to_numpy_array(inventory_cap, expand = True)

        if normalize_env_state:
            assert np.all(inventory_cap != np.inf), "If inventory is normalized, inventory_cap must be a finite number"
        
        # print("start_inventory:", self.start_inventory)
        # print("inventory_cap:", self.inventory_cap)

        # print("set inventory cap")

        if not self.inventory_allowed:
            assert holding_cost is None, "If inventory is not allowed, holding_cost must be None"
        # print("setting holding cost")
        self.holding_cost = self.convert_to_numpy_array(holding_cost, expand = True)

        # print("holding_cost:", self.holding_cost)
        # print("staring setting LT")

        if np.all(lead_time == 0):
            self.use_order_pipeline = False
        else:
            self.use_order_pipeline = True

        self.lead_time = self.convert_to_numpy_array(lead_time, expand = True)
        self.lead_time_variance = self.convert_to_numpy_array(lead_time_variance, expand = True)

        # print("lead_time:", self.lead_time)
        # print("lead_time_variance:", self.lead_time_variance)

        if np.all(lead_time_variance == 0):
            # print("setting LT")
            self.max_lead_time = self.lead_time
            # print("set LT")
        else:
            self.max_lead_time = self.convert_to_numpy_array(max_lead_time, expand = True)

        if order_limit_low is None:
            self.order_limit_low = np.zeros(self.num_products)
            self.order_limit_high = np.ones(self.num_products)
        
        self.order_limit_low = self.convert_to_numpy_array(order_limit_low, expand = True)
        self.order_limit_high = self.convert_to_numpy_array(order_limit_high, expand = True)

        # print("order_limit_low:", self.order_limit_low.shape)
        # print("order_limit_high:", self.order_limit_high.shape)

        order_limit_low = np.repeat(self.order_limit_low, actions_per_product)
        order_limit_high = np.repeat(self.order_limit_high, actions_per_product)

        self.action_space = Box(order_limit_low, order_limit_high, shape=(self.num_products*actions_per_product,))
        # print("action space:", self.action_space.shape)

        # Set observation space
        if observation_space_kwargs is None:
            observation_space_kwargs = {}

        # print("obs kwargs:", observation_space_kwargs)
        self.set_observation_space(**observation_space_kwargs)

        self.calculations = calculations

        # Set MDP info
        mdp_info = MDPInfo(observation_space=self.observation_space,
                           action_space=self.action_space,
                           horizon=horizon,
                           gamma=gamma)
        
        super().__init__(mdp_info)

        self.precision = precision

        self.normalize_reward = normalize_reward

        # print("normalioze rewrad in env class:", self.normalize_reward)
        self.normalize_env_state = normalize_env_state

        # print("done, now resetting")

        self.record_state = record_state

        # Set state variables
        self.reset()

    def set_observation_space(self,
                              features = False,
                              inventory = False,
                              order_pipeline = False,
                              demand_backlog = False
                              ):
        """
        Set observation space. This is used to customize the observation space.

        args:
            features: whether to include features in the observation space
            inventory: whether to include inventory in the observation space
            order_pipeline: whether to include order pipeline in the observation space
            demand_backlog: whether to include demand backlog in the observation space
        """


        features_space = self.num_features if features else 0
        #! for now only focusing on NV problem with features only
        # inventory_space = self.num_products if inventory else 0
        # order_pipeline_space = np.sum(self.num_products * np.max(self.max_lead_time)) if order_pipeline else 0
        # demand_backlog_space = self.num_products if demand_backlog else 0

        # total_obs_space_length = features_space + inventory_space + order_pipeline_space + demand_backlog_space
        total_obs_space_length = features_space

        assert total_obs_space_length > 0, "Need at least one observation dimension"

        low_limit = np.zeros(total_obs_space_length)
        high_limit = np.zeros(total_obs_space_length)

        if features:
            low_limit[:self.num_features] = -np.ones(self.num_features)
            high_limit[:self.num_features] = np.ones(self.num_features)
        else:
            self.num_features = 0

        if inventory:
            low_limit[self.num_features:self.num_features+inventory_space] = np.zeros(inventory_space) # assuming the low inventory end is always 0
            high_limit[self.num_features:self.num_features+inventory_space] = self.inventory_cap

        if order_pipeline:
            slots_occupied = 0
            for product in range(self.num_products):
                # Note: the length of the order pipeline can vary by product. Note that for RNN processing we need to pad the order pipeline to the same length in the preprocessing step
                low_limit[self.num_features+inventory_space+slots_occupied:self.num_features+inventory_space+slots_occupied+self.max_lead_time[product]] = self.order_limit_low[product]
                high_limit[self.num_features+inventory_space+slots_occupied:self.num_features+inventory_space+slots_occupied+self.max_lead_time[product]] = self.order_limit_high[product]
                slots_occupied += self.max_lead_time[product]

        if demand_backlog:
            low_limit[self.num_features+inventory_space+order_pipeline_space:self.num_features+inventory_space+order_pipeline_space+demand_backlog_space] = 0
            high_limit[self.num_features+inventory_space+order_pipeline_space:self.num_features+inventory_space+order_pipeline_space+demand_backlog_space] = self.max_demand_backlog

        # print("features space:", features_space)
        # print("inventory space:", inventory_space)
        # print("order pipeline space:", order_pipeline_space)
        # print("demand backlog space:", demand_backlog_space)
        # print("length:", total_obs_space_length)
        self.observation_space = Box(low_limit, high_limit, shape=(total_obs_space_length,))

    def reset(self, state = None):

        if self.num_observations == self._mdp_info.horizon:
            self.period = 0 # in this environment, the demand will include lag-window observations in the beginning
        else:
            self.period = np.random.choice(self.num_observations-self._mdp_info.horizon)

        self.demand_backlog = np.zeros(self.num_products)
        self.inventory = self.start_inventory.copy()
        self.order_pipeline = np.zeros((self.num_products,np.max(self.max_lead_time)))

        self.set_observation_state()

        return self.observation_state

    def set_observation_state(self):
        
        # TODO: check if this is computationally expensive
        # TODO: make sure this is consistent if some elements are there but not tracked in state

        if self.period >= self.num_observations:
            self.observation_state = np.zeros(self.observation_space.shape)
        
        else:

            if self.dynamic_input is not None:
                starting_point = np.minimum(self.dynamic_input, self.lag_window)
            else:
                starting_point = self.full_additional_context 
            # print("starting_point:", starting_point)
            index = self.period + starting_point
        
            len_history = np.minimum(index, self.lag_window)

            # print("index:", index)
            # print("lag_window:", self.lag_window)
            # print("len_history:", len_history)

            # products on the batch dimension, then time,  then features,
            self.observation_state = np.zeros((self.num_products, len_history, self.observation_space.shape[0]))

            relevant_demand = np.zeros((self.num_products, len_history, len(self.lag_features)))
            # for i in range(len(self.lag_features)):
            #     lag = self.lag_features[i]
            #     relevant_demand[:, :, i] = self.demand[index-len_history-lag+1:index-lag+1, :].T

            for i in range(len(self.lag_features)):
                lag = self.lag_features[i]
                start_index = index - len_history - lag + 1
                end_index = index - lag + 1

                # Check if the start_index is negative
                if start_index < 0:
                    index_smaller_0 = True

                    if end_index >0:
                        end_index_larger_0 = True
                        # Calculate the number of valid data points
                        num_zeros = np.abs(start_index)
                    
                        # Use only the non-negative part for slicing, fill the rest with zeros automatically
                        relevant_demand[:, :num_zeros, i] = 0
                        relevant_demand[:, num_zeros:, i] = self.demand[:end_index, :].T
                    else:
                        end_index_larger_0 = False
                        # If end_index is also negative, fill the entire array with zeros
                        relevant_demand[:, :, i] = 0
                else:
                    index_smaller_0 = False
                    # If start_index is non-negative, proceed as before
                    relevant_demand[:, :, i] = self.demand[start_index:end_index, :].T
            
            self.observation_state[:,:,0:len(self.lag_features)] = relevant_demand

            # check if nan values are present
            if np.isnan(self.observation_state).any():
                print("index below 0:", index_smaller_0)
                print("end index below 0:", end_index_larger_0)
                print(index)
                print(self.period)
                print(self.observation_state)
                assert False

            features_timeless = np.expand_dims(self.features[1], axis=1)
            features_timeless = np.repeat(features_timeless, len_history, axis=1)       

            self.observation_state[:,:,len(self.lag_features):len(self.lag_features)+self.features[1].shape[1]] = features_timeless
            
            relevant_features = self.features[0][index-len_history+1:index+1] # the current observation can be included in the state (unlike in the demand)

            relevant_features_specific = relevant_features[:,:-self.num_products*2] #! TDOD make the 2 variable

            relevant_features_specific_mapped = np.zeros((self.num_products, len_history, self.feature_map.shape[1]))

            #! This is a very slow way to do this. Need to find a better way. Do profile to see if this is a bottleneck
            for i in range(self.num_products):
                # Select the features for this product based on indices_array
                # This uses advanced indexing to pick the correct features
                selected_features = relevant_features_specific[:, self.feature_map[i]]

                # Assign the selected features to the new array, duplicating across the 'history' dimension
                relevant_features_specific_mapped[i] = selected_features

            relevant_features_product = relevant_features[:,-self.num_products*2:] #! TDOD make the 2 variable
            relevant_features_product = relevant_features_product.reshape((len_history, 2, self.num_products)) #! TDOD make the 2 variable
            relevant_features_product = relevant_features_product.transpose((2, 0, 1)) 

            # TODO: ADD relevant_features_specific_mapped and relevant_features_product to the state 
            self.observation_state[:,:,len(self.lag_features)+self.features[1].shape[1]:len(self.lag_features)+self.features[1].shape[1]+relevant_features_specific_mapped.shape[2]] = relevant_features_specific_mapped
            self.observation_state[:,:,len(self.lag_features)+self.features[1].shape[1]+relevant_features_specific_mapped.shape[2]:len(self.lag_features)+self.features[1].shape[1]+relevant_features_specific_mapped.shape[2]+relevant_features_product.shape[2]] = relevant_features_product
          
            #! So far only for features implemented

            if np.isnan(self.observation_state).any():
                print("before rounding")
                print("index below 0:", index_smaller_0)
                print("end index below 0:", end_index_larger_0)
                print(index)
                print(self.period)
                print(self.observation_state)
                assert False
            

            # if self.inventory_allowed:

            #     if self.normalize_env_state:
            #         inv = self.inventory/self.inventory_cap
            #     else:
            #         inv = self.inventory

            #     self.observation_state[self.num_features:self.num_features+self.num_products] = inv

            # if self.use_order_pipeline:
            #     slots_occupied = 0
            #     for product in range(self.num_products):

            #         # print("pipeline_raw:", self.order_pipeline[product])
                    
            #         if self.normalize_env_state:
            #             prod = self.order_pipeline[product]/self.order_limit_high[product]
            #         else:
            #             prod = self.order_pipeline[product]

            #         # print(prod)

            #         self.observation_state[self.num_features+self.num_products+slots_occupied:self.num_features+self.num_products+slots_occupied+self.max_lead_time[product]] = prod
            #         slots_occupied += self.max_lead_time[product]

            # if np.max(self.max_demand_backlog) > 0:
            #     self.observation_state[self.num_features+self.num_products+np.sum(self.num_products * self.lead_time):] = self.demand_backlog

            self.observation_state = np.round(self.observation_state, self.precision)

            # print("inventory:", np.round(self.inventory, 2))
            # print("order pipeline:", np.round(self.order_pipeline, 2))
            # print("_______")

            if np.isnan(self.observation_state).any():
                print("after rounding")
                print("index below 0:", index_smaller_0)
                print("end index below 0:", end_index_larger_0)
                print(index)
                print(self.period)
                print(self.observation_state)
                assert False


    def step(self, action):

        absorbing = False

        reward, info = self.calculations(self, action)

        self.set_observation_state()
        
        # if self.normalize_reward:
        #     print(self.observation_state)

        return self.observation_state, reward, absorbing, info
    
    def render(self):
        pass

    def convert_to_numpy_array(self, value, expand = False):

        """
        Convert a scalar, list, or numpy array to a numpy array.
        
        Parameters:
        - value (scalar, list, numpy.ndarray): The input value.
        
        Returns:
        - numpy.ndarray: The converted numpy array.
        """

        if np.isscalar(value):
            value = np.array([value])
        elif isinstance(value, list):
            value = np.array(value)
        elif isinstance(value, np.ndarray):
            value = value
        elif value is None:
            value = None
        else:
            raise ValueError("Invalid input type. Expected scalar, list, or numpy array.")

        if expand:
            if value is not None:
                if value.size == 1:
                    value = np.repeat(value, self.num_products)
                elif value.size == self.num_products:
                    value = value
                else:
                    raise ValueError("Invalid input size. Expected scalar, list, or numpy array of size 1 or num_products.")
        return value



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../../nbs/agents/rl_agents/pre_specified_agents/11_PPO_classic.ipynb.

# %% auto 0
__all__ = ['PPOClassic']

# %% ../../../../nbs/agents/rl_agents/pre_specified_agents/11_PPO_classic.ipynb 4
# General libraries
import numpy as np

# Torch
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

# Networks
from ...networks.critics import CriticNetworkStandard
from ...networks.actors import ActorNetwork

# Algorithms
from mushroom_rl.algorithms.actor_critic import PPO
from mushroom_rl.policy import GaussianTorchPolicy

# Processors
from ...processors.processors import HybridToContinuous


# %% ../../../../nbs/agents/rl_agents/pre_specified_agents/11_PPO_classic.ipynb 5
class PPOClassic():

    """
    # TODO Update docstring
    """

    def __init__(
            self,
            mdp_info,
            learning_rate_actor = 3e-4,
            learning_rate_critic = None,
            batch_size = 64,
            n_features = 128,
            std_0 = 1.,
            n_epochs_policy = 4,
            eps_ppo = 0.2,
            lam=0.95,
            ent_coeff=0.0,
            n_steps_per_fit=1000,
            optimizer = optim.Adam,
            squeeze_output = True,
            use_cuda = True,
            agent_name = None): 

        self.n_steps_per_fit=n_steps_per_fit
        
        use_cuda = use_cuda and torch.cuda.is_available()
        if use_cuda:
            self.device = "cuda"
        else:
            self.device = "cpu"

        input_shape = mdp_info.observation_space.shape
        print("input shape: ", input_shape)
        actor_output_shape = (mdp_info.action_space.shape[0],) 

        if learning_rate_critic is None:
            learning_rate_critic = learning_rate_actor

        policy_params = dict(network=ActorNetwork, # equivalent to actor params in SAC
                                input_shape=input_shape,
                                output_shape=actor_output_shape,
                                use_cuda=use_cuda,
                                st_0=std_0,
                                n_features=n_features,
                                )

        policy = GaussianTorchPolicy(**policy_params)

        actor_optimizer = {'class': optimizer,
                    'params': {'lr': learning_rate_actor}} 
    
        critic_params = dict(network=CriticNetworkStandard,
                        optimizer={'class': optim.Adam,
                                'params': {'lr': learning_rate_critic}}, 
                        loss=F.mse_loss,
                        n_features=n_features,
                        batch_size = batch_size,
                        input_shape=input_shape,
                        output_shape=(1,),
                        squeeze_output=squeeze_output,
                        use_cuda=use_cuda)
        
        self.agent = PPO(mdp_info, policy, actor_optimizer, critic_params, n_epochs_policy, batch_size, eps_ppo, lam, ent_coeff)
        
        if agent_name is None:
            self.agent.name = 'PPO_classic'
        else:
            self.agent.name = agent_name

    def __getattr__(self, attr):
        return getattr(self.agent, attr)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/agents/benchmark_agents/02_SAA_multi_period_agents.ipynb.

# %% auto 0
__all__ = ['SAA_MP_Agent', 'round_none', 'SAA_MP_Policy', 'WSAA_MP_Agent', 'WSAA_MP_Policy']

# %% ../../../nbs/agents/benchmark_agents/02_SAA_multi_period_agents.ipynb 4
# General libraries:
import numpy as np
import pulp as pl
from tqdm import tqdm
import os
import time

# Mushroom libraries
from mushroom_rl.core import Agent

# ML:
from sklearn.neighbors import KNeighborsRegressor

# %% ../../../nbs/agents/benchmark_agents/02_SAA_multi_period_agents.ipynb 7
class SAA_MP_Agent(Agent):

    train_directly = True
    train_mode = "direct"

    def __init__(self,
                    mdp_info,
                    mdp,
                    h,
                    cu,
                    l,
                    horizon=10,
                    unit_size=0.01,
                    num_scenarios=3,
                    preprocessors = None,
                    postprocessors = None,
                    agent_name = "SAA_MP",
                    precision=5,
                    single_q=False,
                    ):

        self.name = agent_name
        
        policy = SAA_MP_Policy(
            h=h,
            cu=cu,
            l=l,
            horizon=horizon,
            gamma=mdp_info.gamma,
            num_scenarios=num_scenarios,
            mdp = mdp,
            unit_size=unit_size,
            precision=precision,
            preprocessors=preprocessors,
            postprocessors=postprocessors,
            single_q=single_q
        )

        self.precision = precision

        self.train_directly = True
        self.train_mode = "direct"
        self.skip_val = True #! Make true

        super().__init__(mdp_info, policy)

    def fit(self, features = None, demand=None):

        self.policy.get_scenarios(demand)

def round_none(a, b):
    if a is None:
        return None
    else:
        return np.round(a, b)

class SAA_MP_Policy():
    def __init__(self,
        h,
        cu,
        l,
        horizon,
        gamma,
        num_scenarios,
        mdp,
        unit_size,
        precision,
        preprocessors,
        postprocessors,
        single_q,
    ):


        self.h = h
        self.cu = cu
        self.l = l
        self.horizon=horizon+mdp.lead_time[0]
        self.gamma = gamma
        self.num_scenarios = num_scenarios
        self.single_q = single_q

        if preprocessors is None:
            self.preprocessors = list()
        else:
            self.preoprocessors = (preprocessors)
        if postprocessors is None:
            self.postprocessors = list()
        else:
            self.postprocessors = (postprocessors)
        
        self.all_scenarios = np.array([[x for x in range(self.horizon)]])
        self.counter=0
        self.features=False
        self.total_time=0

    def draw_action(self, input):
        for preprocessor in self.preprocessors:
            input = preprocessor(input)
        
        optimization_results = self.optimize(input)
        action = optimization_results['Order']
        
        return np.array([action])

    def optimize(self, state):

        # building_model_start = time.time()
        scenarios, weights = self.sample_scenarios(state)
        if self.features:
            state = {'inventory': state[self.num_features], 'pipeline': state[self.num_features+1:]}
        else:
            state = {'inventory': state[0], 'pipeline': state[1:]}
        cu = self.cu
        l = self.l
        h = self.h
        horizon = self.horizon
        gamma = self.gamma

        T_ = range(horizon)

        # Create the model
        model = pl.LpProblem("Lost_Sales", pl.LpMinimize)

        # Create the decision variables
        ts = []
        for t in T_:
            for s in scenarios.keys():
                ts.append((t,s))

        if self.single_q:
            q = pl.LpVariable.dicts("q", T_, lowBound=0)
        else:
            q = pl.LpVariable.dicts("q", ts, lowBound=0)
        I = pl.LpVariable.dicts("i", ts, lowBound=0)
        u = pl.LpVariable.dicts("u", ts, lowBound=0)

        # Add the objective function
        model += pl.lpSum([(gamma**t)*(weights[s]*(cu*u[t,s] + h*I[t,s])) for t,s in ts])

        # Add the constraints
        I_start = state['inventory']
        p = state['pipeline']
        p = p[::-1]

        # print(scenarios)

        self.start = time.time()

        if self.single_q:
            # for s in scenarios:
            #     model += q[0,s] == q[0,0]
            # print("optimiminzing with single q")

            for s in scenarios:
                # print("scenario", s)
                for t in T_:
                    if t == 0:
                        if l == 0:
                            model += I[t,s] == q[t-l] + I_start  - scenarios[s][t] + u[t,s]
                            model += u[t,s] >= scenarios[s][t] - I_start - q[t-l]
                        else:
                            # print("setting start")
                            model += I[t,s] == p[t] + I_start - scenarios[s][t] + u[t,s]
                            model += u[t,s] >= scenarios[s][t] - I_start - p[t]
                    elif t < l:
                        # print("setting stuff inside the lead time")
                        model += I[t,s] == p[t] + I[t-1,s] - scenarios[s][t] + u[t,s]
                        model += u[t,s] >= scenarios[s][t] - I[t-1,s] - p[t]
                    else:
                        # print("setting further optimization")
                        # print(scenarios[s][t])
                        # print("s:", s)
                        # print("t:", t)
                        # print("q[t-l]:", q[t-l])
                        model += I[t,s] == q[t-l] + I[t-1,s] - scenarios[s][t] + u[t,s]
                        model += u[t,s] >= scenarios[s][t] - I[t-1,s] - q[t-l]
        
        else:
            # print("optimiminzing with q per scenario")
            for s in scenarios:
                model += q[0,s] == q[0,0]
            for s in scenarios:
                for t in T_:
                    if t == 0:
                        if l == 0:
                            model += I[t,s] == q[t-l, s] + I_start  - scenarios[s][t] + u[t,s]
                            model += u[t,s] >= scenarios[s][t] - I_start - q[t-l, s]
                        else:
                            model += I[t,s] == p[t] + I_start - scenarios[s][t] + u[t,s]
                            model += u[t,s] >= scenarios[s][t] - I_start - p[t]
                    elif t < l:
                        model += I[t,s] == p[t] + I[t-1,s] - scenarios[s][t] + u[t,s]
                        model += u[t,s] >= scenarios[s][t] - I[t-1,s] - p[t]
                    else:
                        model += I[t,s] == q[t-l,s] + I[t-1,s] - scenarios[s][t] + u[t,s]
                        model += u[t,s] >= scenarios[s][t] - I[t-1,s] - q[t-l,s]


        # Solve the optimization problem
        # # model.solve(pl.HiGHS_CMD())
        # model.solve()

        # building_model_end = time.time()
        # solving_model_start = time.time()

        # os.environ['GRB_THREADS'] = '1'
        solver = pl.GUROBI_CMD(options=[('Threads', '1')]) # Change this to the solver you are using
        solver.msg = False  # Turn off solver output
        model.solve(solver)
        # Print the results
        # solving_model_end = time.time()
        if self.single_q:
            results = {
                "Cost": sum([weights[s] * (cu*u[t,s].varValue + h*I[t,s].varValue) for t,s in ts]),
                "Order": q[0].varValue,
                }
        else:
            results = {
                "Cost": sum([weights[s] * (cu*u[t,s].varValue + h*I[t,s].varValue) for t,s in ts]),
                "Order": q[0,0].varValue,
                }

        self.end=time.time()

        self.total_time += self.end - self.start
        
        # print(f"Building model: {round_none(building_model_end - building_model_start, 2)}s, Solving model: {round_none(solving_model_end - solving_model_start, 2)}s")
        
        self.counter +=1

        # print(p)
        # print(I_start)
        # for t, s in ts:
        #     if s == 0:
        #         if t<l:
        #             order_minus_l = p[t]
        #         else:
        #             order_minus_l = q[t-l,s].varValue
        #         if t == 0:
        #             I_prev = I_start
        #         else:
        #             I_prev = I[t-1,s].varValue
        #         print(f"{t}, {s}: I_prev: {round_none(I_prev, 2)}, Q_arriving: {round_none(order_minus_l, 2)}, D={scenarios[s][t]}, I={round_none(I[t, s].varValue, 2)}, U={round_none(u[t, s].varValue, 2)}, Q={round_none(q[t, s].varValue, 2)}")
        print(f"\rselected: {results['Order']:.2f} in iteration: {self.counter} with total time: {np.round(self.total_time, 0)}", end="")

        return results
    
    def sample_scenarios(self, state):
        
        scenarios = {}
        weights = {}

        if self.num_scenarios >= self.all_scenarios.shape[0]:
            # Use all available scenarios
            for i, scenario in enumerate(self.all_scenarios):
                scenarios[i] = scenario
                weights[i] = 1 / self.all_scenarios.shape[0]
        
        else:
            # Sample scenarios
            sampled_indices = np.random.choice(self.all_scenarios.shape[0], self.num_scenarios)
            for i, idx in enumerate(sampled_indices):
                scenarios[i] = self.all_scenarios[idx, :]
                weights[i] = 1 / self.num_scenarios

        return scenarios, weights

    def get_scenarios(self, demand):

        self.all_scenarios = np.empty((len(demand) - self.horizon, self.horizon))
        for i in range(len(demand) - self.horizon):
            demand_scenario = np.array(demand[i:i+self.horizon]).squeeze()
            self.all_scenarios[i, :] = demand_scenario

        print("got", self.all_scenarios.shape[0], "scenarios")

    def reset(*args, **kwargs): 
        pass

# %% ../../../nbs/agents/benchmark_agents/02_SAA_multi_period_agents.ipynb 8
class WSAA_MP_Agent(Agent):

    train_directly = True
    train_mode = "direct"

    def __init__(self,
                    mdp_info,
                    mdp,
                    h,
                    cu,
                    l,
                    horizon=10,
                    unit_size=0.01,
                    num_scenarios=3,
                    preprocessors = None,
                    postprocessors = None,
                    agent_name = "SAA_MP",
                    single_q=False,
                    precision=5,
                    method="kNN",
                    method_params=None,
                    ):

        self.name = agent_name
        
        policy = WSAA_MP_Policy(
            num_features = mdp.num_features,
            h=h,
            cu=cu,
            l=l,
            horizon=horizon,
            gamma=mdp_info.gamma,
            num_scenarios=num_scenarios,
            mdp = mdp,
            unit_size=unit_size,
            precision=precision,
            preprocessors=preprocessors,
            postprocessors=postprocessors,
            single_q=single_q,
            method=method,
            method_params = method_params,
        )

        self.precision = precision

        self.train_directly = True
        self.train_mode = "direct"
        self.skip_val = True #! Make true

        super().__init__(mdp_info, policy)

    def fit(self, features = None, demand=None):

        self.policy.get_scenarios(features, demand)

def round_none(a, b):
    if a is None:
        return None
    else:
        return np.round(a, b)

class WSAA_MP_Policy(SAA_MP_Policy):
    def __init__(self,
        num_features,
        h,
        cu,
        l,
        horizon,
        gamma,
        num_scenarios,
        mdp,
        unit_size,
        precision,
        preprocessors,
        postprocessors,
        single_q,
        method="kNN",
        method_params=None,
    ):
        self.method=method
        self.method_params=method_params

        super().__init__(h, cu, l, horizon, gamma, num_scenarios, mdp, unit_size, precision, preprocessors, postprocessors, single_q)

        self.features = True
        self.num_features = num_features
        self.fitted=False
        self.all_scenarios_demand = np.array([[x for x in range(self.horizon)]])
    
    def sample_scenarios(self, state):
        
        scenarios = {}
        weights = {}

        if ((self.num_scenarios >= self.all_scenarios_demand.shape[0]) or (self.fitted==False)):
            # Use all available scenarios
            for i, scenario in enumerate(self.all_scenarios_demand):
                scenarios[i] = scenario
                weights[i] = 1 / self.all_scenarios_demand.shape[0]
        
        else:
            if self.method == "kNN":
                features = np.array(state[:self.num_features]).reshape(1, -1)
                scenarios_ml_dist, scenarios_ml_idx = self.model.kneighbors(features, self.num_scenarios, return_distance=True)
                scenarios_ml_idx = scenarios_ml_idx.squeeze()

            for i in range(self.num_scenarios):
                scenarios[i] = self.all_scenarios_demand[scenarios_ml_idx[i], :]
                weights[i] = 1 / self.num_scenarios

        return scenarios, weights

    def get_scenarios(self, features, demand):

        self.all_scenarios_demand = np.empty((len(demand) - self.horizon, self.horizon))
        for i in range(len(demand) - self.horizon):
            demand_scenario = np.array(demand[i:i+self.horizon]).squeeze()
            self.all_scenarios_demand[i, :] = demand_scenario
        self.all_scenarios_features = np.empty((len(features) - self.horizon, self.horizon, features.shape[1]))
        self.all_scenarios_features = features[:-self.horizon]
        if self.num_scenarios >= self.all_scenarios_demand.shape[0]: 
            print("WARNING: all scenarios are used, no conditional weights are used")
        else:
            if self.method == "kNN":
                self.model = KNeighborsRegressor(n_neighbors=self.num_scenarios, **self.method_params)
                self.model.fit(self.all_scenarios_features, self.all_scenarios_demand)
                self.fitted=True
            else:
                print("method either not recognized or not implemented yet")

        print("got", self.all_scenarios_demand.shape[0], "scenarios")

    def reset(*args, **kwargs): 
        pass
